import requests


HIBP_API_BASE_URI = "https://haveibeenpwned.com/api/v3/"
HIBP_API_ENDPOINT_BREACH_SINGLE = "breach/"
HIBP_API_ENDPOINT_BREACHES = "breaches"
HIBP_API_ENDPOINT_BREACHED_ACCT = "breachedaccount/"
HIBP_API_ENDPOINT_DATA_CLASSES = "dataclasses"
HIBP_API_ENDPOINT_PASTES = "pasteaccount/"


# The headers we send along with each request
pyHIBP_HEADERS = {
    'User-Agent': None,
    'hibp-api-key': None
}


def _require_user_agent(function):
    """
    A decorator which enforces setting the User-Agent on each request. As per the HIBP API, the UA must be set, or a
    HTTP 403 response wil; be raised.

    :raises: RuntimeError if the application User-Agent is not set via ``set_user_agent()`` prior to invoking functions
    which communicate with the HIBP or PwnedPasswords APIs.
    """
    def inner(*args, **kwargs):
        if pyHIBP_HEADERS.get("User-Agent") is None:
            raise RuntimeError("The User-Agent must be set. Call pyhibp.set_user_agent(ua=your_agent_string) first.")
        return function(*args, **kwargs)
    return inner


def set_user_agent(ua: str = None):
    """
    Sets the User-Agent to be used in subsequent calls sent to the HIBP API backend. The UA set should be the name of the
    application implementing the pyhibp module, and accurately describe the nature of the API consumer.

    See: https://haveibeenpwned.com/API/v3#UserAgent

    The HIBP API enforces setting a User-Agent, otherwise an HTTP 403 will be returned.

    :param ua: A string representing the application name which is implementing the pyhibp module which is sent with
    each request to the HIBP API.
    """
    pyHIBP_HEADERS['User-Agent'] = ua


def _require_api_key(function):
    """
    A decorator which enforces setting an API key on functions which require one to be set. Not setting an API key would
    result in an HTTP 401 Unauthorised.

    :raises: RuntimeError if the API key for enabling searching the HIBP API backend by account (email) is not set prior
    to calling functions which invoke said capability.
    """
    def inner(*args, **kwargs):
        if pyHIBP_HEADERS.get('hibp-api-key') is None:
            raise RuntimeError("A HIBP API key is required for this call. Call pyhibp.set_api_key(key=your_key) first.")
        return function(*args, **kwargs)
    return inner


def set_api_key(key: str = None):
    """
    Set an API key for use in all future calls to the pyhibp module which search the HIBP database by email address.

    Per the HIBP API documentation: "Authorisation is required for all APIs that enable searching HIBP by email address,
    namely retrieving all breaches for an account and retrieving all pastes for an account."

    A key can be purchased from <https://haveibeenpwned.com/API/Key>.

    :param key: The key to set for each subsequent request where it is required.
    """
    pyHIBP_HEADERS['hibp-api-key'] = key


def _process_response(response) -> bool:
    """
    Process the `requests` response from the call to the HIBP API endpoints.

    :param response: The response object from a call to `requests`
    :return: True if HTTP Status 200, False if 404. Raises RuntimeError on API-defined status codes of
    400, 403, 429; NotImplementedError if the API returns an unexpected HTTP status code.
    :rtype: bool
    """
    if response.status_code == 200:
        # The request was successful (we found an item)
        return True
    elif response.status_code == 404:
        # The request was successful, though the item wasn't found
        return False
    elif response.status_code == 400:
        # Bad request - The account does not comply with an acceptable format (i.e., it's an empty string)
        raise RuntimeError(
            "HTTP 400 - Bad request - The account does not comply with an acceptable format (i.e., it's an empty string).")
    elif response.status_code == 401:
        # Unauthorised - the API key provided was not valid
        raise RuntimeError(
            "HTTP 401 - Unauthorised - The API key provided was not valid."
        )
    elif response.status_code == 403:
        # Forbidden - no user agent has been specified in the request
        raise RuntimeError("HTTP 403 - User agent required for HIBP API requests, but no user agent was sent to the API endpoint.")
    elif response.status_code == 429:
        # Too many requests - the rate limit has been exceeded
        raise RuntimeError(
            "HTTP 429 - Rate limit exceeded: API rate limit is 1500ms. Retry-After header was: {0}".format(response.headers['Retry-After'])
        )
    else:
        # We /should/ get one of the above error codes. If not, raise an error.
        raise NotImplementedError("Returned HTTP status code of {0} was not expected.".format(response.status_code))


@_require_api_key
@_require_user_agent
def get_account_breaches(
        account: str = None,
        domain: str = None,
        truncate_response: bool = False,
        include_unverified: bool = False,
        timeout=5.0) -> list:
    """
    Gets breaches for a specified account from the HIBP system, optionally restricting the returned results
    to a specified domain.

    This function requires a HIBP API key to be set. See ``set_api_key()``.

    :param account: The user's account name (such as an email address or a user-name). Default None. `str` type. Required.
    :param domain: The domain to check for breaches. Default None. `str` type. Optional
    :param truncate_response: If ``account`` is specified, truncates the response down to the breach names.
    Default False. `bool` type.
    :param include_unverified: If set to True, unverified breaches are included in the result. Default False. `bool` type
    :param timeout: The timeout value to be passed to the underlying `requests.get()` call. Default 5.0.
    :return: A list object containing one or more dict objects, based on the information being requested,
    provided there was matching information. Boolean False returned if no information was found according to
    the HIBP API.
    :rtype: list
    """
    if account is None or not isinstance(account, str):
        raise AttributeError("The account parameter must be specified, and must be a string.")
    # The domain does not need to specified, but is must be text, if so.
    if domain is not None and not isinstance(domain, str):
        raise AttributeError("The domain parameter, if specified, must be a string.")

    uri = HIBP_API_BASE_URI + HIBP_API_ENDPOINT_BREACHED_ACCT + account

    # Build the query string payload (requests drops params when None)
    # (and the HIBP backend ignores those that don't apply)
    query_string_payload = {
        "domain": domain,
        "truncateResponse": truncate_response,
        "includeUnverified": include_unverified,
    }
    resp = requests.get(url=uri, params=query_string_payload, headers=pyHIBP_HEADERS, timeout=timeout)

    if _process_response(response=resp):
        return resp.json()
    else:
        return []


@_require_user_agent
def get_all_breaches(domain: str = None, timeout=5.0) -> list:
    """
    Returns a listing of all sites breached in the HIBP database.

    :param domain: Optional, default None. If specified, get all breaches for the domain with the specified name. `str` type.
    :return: A list object containing one or more dict objects if breaches are present. Returns Boolean False.
    :param timeout: The timeout value to be passed to the underlying `requests.get()` call. Default 5.0.
    if ``domain`` is specified, but the resultant list would be length zero.
    :rtype: list
    """
    if domain is not None and not isinstance(domain, str):
        raise AttributeError("The domain parameter, if specified, must be a string.")

    uri = HIBP_API_BASE_URI + HIBP_API_ENDPOINT_BREACHES
    query_string_payload = {'domain': domain}
    resp = requests.get(url=uri, params=query_string_payload, headers=pyHIBP_HEADERS, timeout=timeout)

    # The API will return HTTP200 even if resp.json is length zero.
    if _process_response(response=resp) and len(resp.json()) > 0:
        return resp.json()
    else:
        return []


@_require_user_agent
def get_single_breach(breach_name: str = None, timeout=5.0) -> dict:
    """
    Returns a single breach's information from the HIBP's database.

    :param breach_name: The breach to retrieve. Required. `str` type.
    :param timeout: The timeout value to be passed to the underlying `requests.get()` call. Default 5.0.
    :return: A dict object containing the information for the specified breach name, if it exists in the HIBP
    database. Boolean False is returned if the specified breach was not found.
    :rtype: dict
    """
    if not isinstance(breach_name, str):
        raise AttributeError("The breach_name must be specified, and be a string.")

    uri = HIBP_API_BASE_URI + HIBP_API_ENDPOINT_BREACH_SINGLE + breach_name
    resp = requests.get(url=uri, headers=pyHIBP_HEADERS, timeout=timeout)

    if _process_response(response=resp):
        return resp.json()
    else:
        return {}


@_require_api_key
@_require_user_agent
def get_pastes(email_address: str = None, timeout=5.0) -> list:
    """
    Retrieve all pastes for a specified email address.

    This function requires a HIBP API key to be set. See ``set_api_key()``.

    :param email_address: The email address to search. Required. `str` type.
    :param timeout: The timeout value to be passed to the underlying `requests.get()` call. Default 5.0.
    :return: A list object containing one or more dict objects corresponding to the pastes the specified email
    address was found in. Boolean False returned if no pastes are detected for the given account.
    :rtype: list
    """
    if not isinstance(email_address, str):
        raise AttributeError("The email address supplied must be provided, and be a string.")

    uri = HIBP_API_BASE_URI + HIBP_API_ENDPOINT_PASTES + email_address
    resp = requests.get(url=uri, headers=pyHIBP_HEADERS, timeout=timeout)

    if _process_response(response=resp):
        return resp.json()
    else:
        return []


@_require_user_agent
def get_data_classes(timeout=5.0) -> list:
    """
    Retrieves all available data classes from the HIBP API.

    :param timeout: The timeout value to be passed to the underlying `requests.get()` call. Default 5.0.
    :return: A list object containing available data classes, corresponding to attributes found in breaches.
    A given breach will have one or more of the data classes in the list.
    :rtype: list
    """
    uri = HIBP_API_BASE_URI + HIBP_API_ENDPOINT_DATA_CLASSES
    resp = requests.get(url=uri, headers=pyHIBP_HEADERS, timeout=timeout)

    if _process_response(response=resp):
        return resp.json()
    else:
        # This path really shouldn't return false
        raise RuntimeError("HIBP API returned HTTP404 on a request for data classes.")



import hashlib

import requests

from pyhibp import _require_user_agent, pyHIBP_HEADERS

PWNED_PASSWORDS_API_BASE_URI = "https://api.pwnedpasswords.com/"
PWNED_PASSWORDS_API_ENDPOINT_RANGE_SEARCH = "range/"

RESPONSE_ENCODING = "utf-8-sig"


def is_password_breached(password: str = None, sha1_hash: str = None, add_padding: bool = False, timeout=5.0) -> int:
    """
    Execute a search for a password via the k-anonymity model, checking for hashes which match a specified
    prefix instead of supplying the full hash to the Pwned Passwords API.

    Uses the first five characters of a SHA-1 hash to provide a list of hash suffixes along with the
    number of times that hash appears in the data set. In doing so, the API is not provided the information
    required to reconstruct the password (e.g., by brute-forcing the hash).

    Either ```password`` or ``sha1_hash`` must be specified. Only one parameter should be provided.

    The precedence of parameters is as follows:
    1) password - Used to compute the SHA-1 hash of the password.
    2) sha1_hash - The hash prefix (hash[0:5]) is passed to the HIBP API, and this function will check the returned list of
    hash suffixes to determine if a breached password was in the HIBP database.

    :param password: The password to check. Will be converted to a SHA-1 string. `str` type.
    :param sha1_hash: A full SHA-1 hash. `str` type.
    :param add_padding: Whether padding should be used when performing the check (obfuscates response size, does not
    alter return type/value.
    :param timeout: The timeout value to be passed to the underlying `requests.get()` call. Default 5.0.
    :return: An Integer representing the number of times the password is in the data set; if not found,
    Integer zero (0) is returned.
    :rtype: int
    """
    # Parameter validation section
    if not any([password, sha1_hash]):
        raise AttributeError("Either password or sha1_hash must be provided.")
    elif password is not None and not isinstance(password, str):
        raise AttributeError("password must be a string.")
    elif sha1_hash is not None and not isinstance(sha1_hash, str):
        raise AttributeError("sha1_hash must be a string.")

    if password:
        sha1_hash = hashlib.sha1(password.encode('utf-8')).hexdigest()
    if sha1_hash:
        # The HIBP API stores the SHA-1 hashes in uppercase, so ensure we have it as uppercase here
        sha1_hash = sha1_hash.upper()
        hash_prefix = sha1_hash[0:5]

    suffix_list = suffix_search(hash_prefix=hash_prefix, add_padding=add_padding, timeout=timeout)

    # Since the full SHA-1 hash was provided, check to see if it was in the resultant hash suffixes returned.
    for hash_suffix in suffix_list:
        if sha1_hash[5:] in hash_suffix:
            # We found the full hash, so return
            return int(hash_suffix.split(':')[1])

    # If we get here, there was no match to the supplied SHA-1 hash; return zero.
    return 0


@_require_user_agent
def suffix_search(hash_prefix: str = None, add_padding: bool = False, timeout=5.0) -> list:
    """
    Returns a list of SHA-1 hash suffixes, consisting of the SHA-1 hash characters after position five,
    and the number of times that password hash was found in the HIBP database, colon separated.

    Leveraging the k-Anonymity model, a list of hashes matching a specified SHA-1 prefix are returned.
    From this list, the calling application may determine if a given password was breached by comparing
    the remainder of the SHA-1 hash against the returned results. As an example, for the hash prefix of
    '42042', the hash suffixes would be:

    ```
    005F4A4B9265A2BABE10B1A9AB9409EA3F0:1
    00D6F0319225107BD5736B72717BD381660:8
    01355DCE0B54F0E8DBBBA8F7B9A9872858A:15
    0163E1C872A64A62625F5EB2F3807B7F90B:2
    020DDE278E6A9C05B356C929F254CE6AED5:1
    021EFB4FAE348050D9EDCD10F8B6A87C957:4
    ...
    ```

    If the `prefix` and `suffix` form a complete SHA-1 hash for the password being compared, then it
    indicates the password has been found in the HIBP database.

    :param add_padding: Boolean. Adds padding to the response to include hash suffixes which have not been breached, in
    order to prevent sniffing of response size to infer what hash prefix was searched. Entries which end in zero can be
    disregarded.
    :param hash_prefix: The first five characters of a SHA-1 hash. `str` type.
    :param timeout: The timeout value to be passed to the underlying `requests.get()` call. Default 5.0.
    :return: A list of hash suffixes.
    :rtype: list
    """
    if not hash_prefix or not isinstance(hash_prefix, str):
        raise AttributeError("hash_prefix must be a supplied, and be a string.")
    if hash_prefix and len(hash_prefix) != 5:
        raise AttributeError("hash_prefix must be of length 5.")

    uri = PWNED_PASSWORDS_API_BASE_URI + PWNED_PASSWORDS_API_ENDPOINT_RANGE_SEARCH + hash_prefix

    _headers = pyHIBP_HEADERS
    _headers['Add-Padding'] = "true" if add_padding else None

    resp = requests.get(url=uri, headers=_headers, timeout=timeout)
    if resp.status_code != 200:
        # The HTTP Status should always be 200 for this request
        raise RuntimeError("Response from the endpoint was not HTTP200; this should not happen. Code was: {0}".format(resp.status_code))
    # The server response will have a BOM if we don't do this.
    resp.encoding = RESPONSE_ENCODING

    return resp.text.split()
